﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Operators;
using HeuristicLab.Optimization;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.PluginInfrastructure;

namespace HeuristicLab.Problems.GrammaticalOptimization {
  // this type is not storable because we use a func<ITree,double> for evaluation, which references back to the original grammatical optimization problem
  [NonDiscoverableType]
  [Item("GrammaticalOptimizationEvaluator", "Evaluator for grammatical optimization problems (using a symbolic expression tree encoding).")]
  public class GrammaticalOptimizationEvaluator : SingleSuccessorOperator, IGrammaticalOptimizationEvaluator {
    public event Action<string, double> SolutionEvaluated;

    public ILookupParameter<DoubleValue> QualityParameter {
      get { return (ILookupParameter<DoubleValue>)Parameters["Quality"]; }
    }
    public ILookupParameter<ISymbolicExpressionTree> SymbolicExpressionTreeParameter {
      get { return (ILookupParameter<ISymbolicExpressionTree>)Parameters["SymbolicExpressionTree"]; }
    }

    private Func<string, double> evalFunc;
    private Func<ISymbolicExpressionTree, string> toStringFunc;

    public GrammaticalOptimizationEvaluator(GrammaticalOptimizationEvaluator original, Cloner cloner)
      : base(original, cloner) {
      this.evalFunc = original.evalFunc;
    }

    public GrammaticalOptimizationEvaluator(Func<ISymbolicExpressionTree, string> toStringFunc, Func<string, double> evalFunc)
      : base() {
      Parameters.Add(new LookupParameter<DoubleValue>("Quality"));
      Parameters.Add(new LookupParameter<ISymbolicExpressionTree>("SymbolicExpressionTree"));
      this.evalFunc = evalFunc;
      this.toStringFunc = toStringFunc;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new GrammaticalOptimizationEvaluator(this, cloner);
    }

    public override IOperation Apply() {
      var tree = SymbolicExpressionTreeParameter.ActualValue;
      var sentence = toStringFunc(tree);
      var q = evalFunc(sentence);
      QualityParameter.ActualValue = new DoubleValue(q);

      RaiseSolutionEvaluated(sentence, q);

      return base.Apply();
    }

    private void RaiseSolutionEvaluated(string sentence, double quality) {
      var handler = SolutionEvaluated;
      if (handler != null) handler(sentence, quality);
    }
  }
}
